(* TENLAB COMPILATION Module by Mehmet Kerem Turkcan and Yusuf Cem Subakan *)
(* Based on the corresponding MICROC Module                                *)

open Ccode
open Str

module StringMap = Map.Make(String)
(*
 This converts the dimension matching list into a binary list 
*)
let rec list2bin lst dims cnt =
  match lst,dims with 
    | [],[] -> []
    | _,[] -> []
    | [],hd::tl2 -> 0::(list2bin [] tl2 (cnt+1))
    | hd::tl,hd2::tl2 -> if (cnt=hd) then 1::(list2bin tl tl2 (cnt+1)) 
        else 0::(list2bin lst tl2 (cnt+1));;
(*
This function fills the unmatched indices of a tensor in a list
*) 
let rec binary2inds match_dims len charac = 
  match match_dims with 
    | [] -> []
    | hd::tl -> if (hd=0) then (charac^(string_of_int ( len - 
        (List.length tl))))::(binary2inds tl len charac)
        else binary2inds tl len charac;;
(*
This function extracts the dimension limits for specified limits in the binary 
list binsA / one_or_zero input specifies whether matched/non-matched indices 
should be picked
*)
let rec binary2lims binsA dims one_or_zero = 
  match binsA,dims with 
    | [],[] -> []
    | _,[] -> []
    | [],_ -> [] 
    | hd::tl,hd2::tl2 -> if (hd=one_or_zero) then 
        hd2::(binary2lims tl tl2 one_or_zero) 
        else binary2lims tl tl2 one_or_zero;;  
(*
This functions writes the for loops, for the observable indices of a tensor in a list
*) 
let rec writefors inds dims = 
  match inds,dims with 
    | [],[] -> []
    | hd::tl,[] -> []
    | [],hd2::tl2 -> [] 
    | hd::tl,hd2::tl2 -> ("for(int "^hd^"=0;"^hd^"<"^(string_of_int hd2)^
        ";"^hd^"++) { \n " )::(writefors tl tl2);;
(*
This produces a list of length len consisting of all n's
*)
let rec all_n_list n len cnt = 
  if (cnt<=len) then n::(all_n_list n len (cnt+1))
  else [];;         
(*
This function is needed for linear indexing
*)
let rec multiply_lims lims ind cnt = 
  match lims with 
    | [] -> ""
    | hd::tl -> if (cnt<ind-1) then (string_of_int hd) ^ "*" ^
        (multiply_lims tl ind (cnt+1)) 
        else if (cnt = ind-1) then string_of_int hd 
        else "";;
(*
This function writes the linear indices given the index and dimensions list
*)
let rec linear_indices inds lims len cnt= 
  match inds with 
    | [] -> ""
    | hd::tl -> if (cnt = 1 && len > 1) then hd ^ "+" ^ 
        (linear_indices tl lims len (cnt+1))
        else if (cnt = 1 && len = 1) then hd  
        else if (cnt > 1 && cnt<len && len>1) then 
        hd^"*"^(multiply_lims lims cnt 1) ^ "+" ^
        (linear_indices tl lims len (cnt+1))
        else hd^"*"^(multiply_lims lims cnt 1) ^ 
        (linear_indices tl lims len (cnt+1));;
(*
This function negates a binary list
*) 
let rec negate_list lst = 
  match lst with 
  | [] -> []
  | hd::tl -> (1-hd)::(negate_list tl);;

        
(*
This function forms the indices/limits given the binary matching list binsA, 
observable indices obsinds and matched indices matchedinds
*) 
let rec form_indsA binsA obsinds matchedinds cnt1 cnt2 = 
  match binsA with 
    | [] -> [] 
    | hd::tl -> if (hd=0) then (List.nth obsinds cnt1)::
      (form_indsA tl obsinds matchedinds (cnt1+1) cnt2)
      else (List.nth matchedinds cnt2)::
      (form_indsA tl obsinds matchedinds cnt1 (cnt2+1));;  
(*
This function finds the index of the element x in list lst - 
the index starts from 1
*)         
let find_index lst x = 
  let rec find_x_index lst x cnt =     
    match lst with
      | [] ->  -1 
      | hd::tl -> if (hd=x) then cnt 
        else find_x_index tl x (cnt+1) 
        in find_x_index lst x 0;; 
(*
This function forms the indices of B
*)
let rec form_indsB binsB obsinds matchedinds matchB = 
  let rec form binsB obsinds matchedinds matchB cnt1 cnt2 =
    match binsB with 
    | [] -> []
    | hd::tl -> if (hd=0) then 
    (List.nth obsinds cnt1)::
    (form tl obsinds matchedinds matchB (cnt1+1) (cnt2+1))
      else (List.nth matchedinds (find_index matchB cnt2))::
      (form tl obsinds matchedinds matchB cnt1 (cnt2+1)) 
        in form binsB obsinds matchedinds matchB 0 1;;                    

let rec concat_strings_inlist lst = 
        match lst with 
        | [] -> ""
        | hd::tl -> hd^(concat_strings_inlist tl);;

(*
Write the 'observable' for loops for tensor A in a list called forsA
*) 
let create_tensor_product_code c sum_or_diag dimsA a matchA dimsB b matchB =  
  let binsA = list2bin matchA dimsA 1 in
  let indsA = binary2inds binsA (List.length binsA) "TENLAB_i" in
  let limsA = binary2lims binsA dimsA 0 in
  let forsA = writefors indsA limsA in

  (*
  Write the 'observable' for loops for tensor B ins a list called forsB
  *)
  let binsB = list2bin (List.sort compare matchB) dimsB 1 in 
  let indsB = binary2inds binsB (List.length binsB) "TENLAB_j" in 
  let limsB = binary2lims binsB dimsB 0 in 
  let forsB = writefors indsB limsB in 

  (*
  Write the 'matched' for loops 
  *)
  let inds_matched_all = binary2inds 
    (*
    Get inds
    *)
    (all_n_list 0 (List.length sum_or_diag) 1 ) (List.length sum_or_diag) 
    "TENLAB_k" in
  let inds_matched_observable = 
    binary2inds sum_or_diag (List.length sum_or_diag) "TENLAB_k" in
  let inds_matched_collapsed = 
    binary2inds (negate_list sum_or_diag) (List.length sum_or_diag) "TENLAB_k" in
  (*
  Get the limits
  *)
  let lims_matched_all = binary2lims binsA dimsA 1 in
  let lims_matched_observable = binary2lims sum_or_diag lims_matched_all 0 in
  let lims_matched_collapsed = binary2lims sum_or_diag lims_matched_all 1 in
  (*
  Write for's
  *) 
  let fors_matched_observable = 
    writefors inds_matched_observable lims_matched_observable in
  let fors_matched_collapsed = 
    writefors inds_matched_collapsed lims_matched_collapsed in
  (*
  Concatanete to get all fors
  *)
  let all_fors = forsA @ forsB @ fors_matched_observable @ 
    fors_matched_collapsed in
  (*
  get the indices and limit of the term C by concatenating 
  observable indices/limits
  *)
  let all_inds_C = indsA @ ( indsB @ inds_matched_observable) in
  let all_lims_C = limsA @ ( limsB @ lims_matched_observable) in
  (*
  write down the linear indices for C
  *)
  let term_C = linear_indices all_inds_C all_lims_C 
    (List.length all_inds_C) 1 in
  (*
  Get all indices for A
  *)
  let all_inds_A = form_indsA binsA indsA inds_matched_all 0 0 in
  (*
  Write down what we have for A in linear indices
  *)
  let term_A = linear_indices all_inds_A dimsA (List.length all_inds_A) 1 in

  let all_inds_B = form_indsB binsB indsB inds_matched_all matchB in
  (*
  Write down what we have for B in linear indices
  *)
  let term_B = linear_indices all_inds_B dimsB (List.length all_inds_B) 1 in
  (*
  Aggregate the terms inside fors
  *) 
  let theterm_inside_fors = c^".Content["^term_C^"] = " ^
    c ^ ".Content[" ^ term_C ^ "] + " ^ a ^ ".Content[" ^ term_A ^ "] * " ^
    b ^ ".Content[" ^ term_B ^ "]; \n" in
  (*
  Form the closing brackets
  *)
  let closing_brackets = all_n_list "} \n " ((List.length all_fors)-1) 0 in 
  (*
  Put everything inside a big string - the output dimensions are in all_lims_C
  *)
  let everything = 
    (concat_strings_inlist all_fors) ^ theterm_inside_fors ^ 
    (concat_strings_inlist closing_brackets) in everything;;



let str_crop_last_char x_in =
  if x_in = "" then "" else
  String.sub x_in 0 ((String.length x_in) - 1)
  
let get_string_length x_in = string_of_int((( String.length x_in)-1)/2);;

let get_tensor_dimension e =
  let temp = List.filter (fun x -> (String.length x)>0) (List.map (function
    | Str.Delim s -> s
    | _ -> "") e) in
  let mapper x_in=  ((String.length x_in)-1)/2 in
  let temp2 = List.sort_uniq (fun x y -> if x > y then 1 else 0) 
    (List.map (mapper) temp) in
  temp2;;
	
let get_tensor_blocks x_in = 
  Str.full_split (Str.regexp "\\[\\([0-9]+[',']\\)+[0-9]+\\]") x_in;;
	
let rec build_tensor e =
  if (List.length e) > 1 then
  let temp = (List.map (function
  				         | Str.Delim s -> (get_string_length s)
  				         | Str.Text  s ->  s) e) in
  let temp1 = build_tensor (get_tensor_blocks (String.concat "" temp)) in
  let temp2 = get_tensor_dimension e in
  temp2::temp1 else
  [];;
	
let get_tensor_elements x_in = 
  List.map int_of_string (Str.split (Str.regexp "[^0-9]+") x_in);;
	

let declare_tensor_in_C name_in content_in= 
  let temp1 = (List.concat (build_tensor 
    (get_tensor_blocks ("[" ^ content_in ^ "]")))) in
  let temp3 = List.fold_left (fun x1 x2 -> x1 ^ x2 ^ ");\nTENLAB_assign(&" ^ 
    name_in ^ ",") ("TENLAB_assign(&" ^ name_in ^ ",") 
    ( (Str.split (Str.regexp "[^0-9]+") content_in)) in
  let temp4 = String.sub temp3 0 
    ((String.length temp3)-17-(String.length name_in)) in
  let temp5 = List.fold_left (fun x1 x2 -> x1 ^ (string_of_int x2) ^ 
    ");\nTENLAB_add_shape(&" ^ name_in ^ ",") 
    ("\nTENLAB_add_shape(&" ^ name_in ^ ",") temp1 in
  let temp6 = String.sub temp5 0 
    ((String.length temp5)-19-(String.length name_in)) in
  "TENLAB_Tensor " ^ name_in ^ 
  ";\nTENLAB_Tensor_create(&" ^ name_in ^ ");\n" ^ temp4 ^ temp6;;
	
let declare_tensor_in_C_without_first_line name_in content_in= 
  let temp1 = (List.concat (build_tensor 
    (get_tensor_blocks ("[" ^ content_in ^ "]")))) in
  let temp3 = List.fold_left (fun x1 x2 -> x1 ^ x2 ^ ");\nTENLAB_assign(&" ^ 
    name_in ^ ",") ("TENLAB_assign(&" ^ name_in ^ ",") 
    ( (Str.split (Str.regexp "[^0-9]+") content_in)) in
  let temp4 = String.sub temp3 0 
    ((String.length temp3)-17-(String.length name_in)) in
  let temp5 = List.fold_left (fun x1 x2 -> x1 ^ (string_of_int x2) ^ 
    ");\nTENLAB_add_shape(&" ^ name_in ^ ",") 
    ("\nTENLAB_add_shape(&" ^ name_in ^ ",") temp1 in
  let temp6 = String.sub temp5 0 
    ((String.length temp5)-19-(String.length name_in)) in
  [DirectOut ((*"TENLAB_Tensor_create(&" ^ name_in ^ ");\n" ^*) temp4 ^ temp6)];;
  
(* 
  "double *" ^ name_in ^ ";\n" ^ name_in ^ " = 
  (double["  ^ string_of_int temp2 ^ "]) {" ^ temp4 ^ "};\nint *TENLAB_" ^ 
  name_in ^"_size;\nTENLAB_" ^ name_in ^ "_size = 
  (int[" ^ string_of_int (List.length temp1) ^ "]) {" ^ temp6 ^ "};\n" 
*)

(* Temporary Statement Constructors: *)

(*let declare_tensor_in_C name_in content_in = 
	let temp1 = (List.concat (build_tensor (get_tensor_blocks 
	  ("[" ^ content_in ^ "]")))) in
	let temp2 = (List.fold_left (fun x1 x2 -> x1 * x2) 1 temp1) in
	let temp3 = List.fold_left (fun x1 x2 -> x1 ^ "," ^ x2) "" 
	  ((Str.split (Str.regexp "[^0-9]+") content_in)) in
	let temp4 = String.sub temp3 1 ((String.length temp3)-1) in
	let temp5 = List.fold_left (fun x1 x2 -> x1 ^ "," ^ 
	  (string_of_int x2)) "" temp1 in
	let temp6 = String.sub temp5 1 ((String.length temp5)-1) in
	"double *" ^ name_in ^ ";\n" ^ name_in ^ " = 
	(double["  ^ string_of_int temp2 ^ "]) {" ^ temp4 ^ "};\nint *TENLAB_" ^ 
	name_in ^"_size;\nTENLAB_" ^ name_in ^ "_size = (int[" ^ string_of_int 
	(List.length temp1) ^ "]) {" ^ temp6 ^ "};\n";;*)
	
	
(* Deprecated: Memory Management via Compound Literals: For Future Use *)

(* 
let declare_tensor_in_C name_in content_in = 
	let temp1 = (List.concat (build_tensor 
	  (get_tensor_blocks ("[" ^ content_in ^ "]")))) in
	let temp2 = (List.fold_left (fun x1 x2 -> x1 * x2) 1 temp1) in
	let temp3 = List.fold_left (fun x1 x2 -> x1 ^ "," ^ x2) "" 
	  ( (Str.split (Str.regexp "[^0-9]+") content_in)) in
	let temp4 = String.sub temp3 1 ((String.length temp3)-1) in
	let temp5 = List.fold_left 
	  (fun x1 x2 -> x1 ^ "," ^ (string_of_int x2)) "" temp1 in
	let temp6 = String.sub temp5 1 ((String.length temp5)-1) in
	"double *" ^ name_in ^ ";\n" ^ name_in ^ 
	  " = (double["  ^ string_of_int temp2 ^ 
	  "]) {" ^ temp4 ^ "};\nint *TENLAB_" ^ name_in ^"_size;\nTENLAB_" ^ 
	  name_in ^ "_size = (int[" ^ 
	  string_of_int (List.length temp1) ^ "]) {" ^ temp6 ^ "};\n";;
	
let declare_tensor_in_C_without_first_line name_in content_in = 
	let temp1 = (List.concat (build_tensor 
	  (get_tensor_blocks ("[" ^ content_in ^ "]")))) in
	let temp2 = (List.fold_left (fun x1 x2 -> x1 * x2) 1 temp1) in
	let temp3 = List.fold_left (fun x1 x2 -> x1 ^ "," ^ x2) "" 
	  ( (Str.split (Str.regexp "[^0-9]+") content_in)) in
	let temp4 = String.sub temp3 1 ((String.length temp3)-1) in
	let temp5 = List.fold_left (fun x1 x2 -> x1 ^ "," ^ 
	  (string_of_int x2)) "" temp1 in
	let temp6 = String.sub temp5 1 ((String.length temp5)-1) in
	let output_string = name_in ^ " = (double["  ^ 
	  string_of_int temp2 ^ "]) {" ^ temp4 ^ "};\nTENLAB_" ^ 
	  name_in ^ "_size = (int[" ^ string_of_int (List.length temp1) ^ 
	  "]) {" ^ temp6 ^ "};\n" in
	[DirectOut output_string];;
*)

let declare_for_loop_in_C name_a_in name_b_in = 
	"int TENLAB_for_" ^ name_a_in ^ ";\nfor(TENLAB_for_" ^ name_a_in ^ 
	"=1;TENLAB_for_" ^ name_a_in ^ "<=" ^ name_b_in ^ 
	".cur_content_length;TENLAB_for_" ^ name_a_in ^ "++){\n";;

let declare_arbitrary_for_loop_in_C name_a_in name_b_in = 
	[DirectOut ("int TENLAB_for_" ^ name_a_in ^ ";\nfor(TENLAB_for_" ^ 
	name_a_in ^ "=1;TENLAB_for_"^ name_a_in ^"<=" )] @ name_b_in @ 
	[DirectOut (";TENLAB_for_" ^ name_a_in ^ "++){\n")];;

	
let list_dot_product = List.fold_left2 (fun s x y -> s + x * y) 0;;

(*let tensor_operation_abstract e1 e2 op_type = [TensorOpt [e1,e2,op_type]];;*)

let tensor_operation_constructor e1 e2 e3 op_type= 
  [DirectOut (declare_tensor_in_C "TENLAB_temp_sum_a" e2)] @ 
  [DirectOut (declare_tensor_in_C "TENLAB_temp_sum_b" e3)] @ 
  [DirectOut ("double *" ^ e1 ^ ";")] @ 
  [DirectOut (declare_for_loop_in_C"TENLAB_inner"  "TENLAB_temp_sum_a")] @ 
  [DirectOut ("*" ^ e1 ^ 
  "[TENLAB_for_TENLAB_inner] = TENLAB_temp_sum_a[TENLAB_for_TENLAB_inner] " ^ 
  op_type ^ " TENLAB_temp_sum_b[TENLAB_for_TENLAB_inner];")] @ [Curlend];;


(* let get_single_element_from_tensor_in_C name_a_in name_b_in =
	let temp1 = get_tensor_elements name_b_in in 
	let temp2 = in
	"double " ^ name_a_in ^ ";\n" ^ name_a_in ^ "=" ^   *)
	
(* For Debugging:
get_tensor_elements "[[[5,4,3,7],[1,2,3,4]]]";;
get_tensor_blocks "[[[5,4,3,7],[1,2,3,4]]]";;
let result = build_tensor (get_tensor_blocks "[[[5,4,3,7],[1,2,3,4]]]") in 
  (List.concat result);;
*)

(* 
Alternate Parser Module: 
tensor_lists:
  tensor_list                   { $1 }
    | tensor_lists COMMA tensor_list { (($1)) ^ "," ^ (($3)) }
    | LNPAREN tensor_lists RNPAREN { "[" ^ (($2))  ^ "]" }
  
tensor_list:
  LNPAREN tensor_element_list RNPAREN  { "[" ^ (String.concat "" ($2)) ^ "]" }
  
tensor_element_list:
  LITERAL                    { [(string_of_int $1)] }
  tensor_element_list COMMA LITERAL 
    { List.rev ((string_of_int $3) :: ("," :: $1)) } 
*)

(* 
Symbol table: Information about all the names in scope 
*)
type env = {
    function_index : int StringMap.t;
    global_index   : int StringMap.t; 
    local_index    : int StringMap.t; 
}

let rec enum stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs


(*  
Translate a program in AST form into a C program in blocks.  Throw an exception 
if something is wrong, e.g., a reference to an unknown variable or function
*)
let translate (globals, functions) main_type=

  (* Allocate "addresses" for each global variable *)
  let global_indexes = string_map_pairs StringMap.empty (enum 1 0 globals) in

  (* 
  Assign indexes to special function names; built-in "print" is special 
  *)
  let built_in_functions = StringMap.add "print" (-1) StringMap.empty in
  let built_in_functions = StringMap.add "input" (-2) built_in_functions in
  let built_in_functions = StringMap.add "output" (-3) built_in_functions in
  let built_in_functions = StringMap.add "pop" (-4) built_in_functions in
  let built_in_functions = StringMap.add "dequeue" (-5) built_in_functions in
  let built_in_functions = StringMap.add "reshape" (-6) built_in_functions in
  let built_in_functions = StringMap.add "shape" (-7) built_in_functions in
  let built_in_functions = StringMap.add "set" (-8) built_in_functions in
  let built_in_functions = StringMap.add "length" (-9) built_in_functions in
  let function_indexes = string_map_pairs built_in_functions
    (enum 1 1 (List.map (fun f -> f.fname) functions)) in

  (* 
  Translate into C, keeping track of the edge cases 
  *)
  let translate env fdecl =
  (*
	Bookkeeping: Get number of inputs 
	*)
    let num_formals = List.length fdecl.formals
    and local_offsets = enum 1 1 fdecl.locals
    and formal_offsets = enum (-1) (-2) fdecl.formals in
    let env = { env with local_index = string_map_pairs
		        StringMap.empty (local_offsets @ formal_offsets) } in
	(* 
	Reorder the Assignment Expressions in Intermediate Representation via
	Temporary Variables and Generate the C Code 
	*)
	let expr_to_tensor_op(listin)=
	  let rec env_build(listin, temporary_declarations, construction,
	    finaldeclaration,tempmemorycollector,declarationtype,
	    declarationsize,memoryimprint,varname)=
		  match listin with
		    | [] -> if declarationtype = 1 then  
			    (List.rev temporary_declarations) @ 
			    (memoryimprint) @ 
			    (List.rev tempmemorycollector) else 
			    (List.rev temporary_declarations) @ 
          [DirectOut "TENLAB_assign(&"] @ 
			    (List.rev finaldeclaration) @ [DirectOut ","] @ 
			    (List.rev construction) @ [DirectOut ")"] @ 
			    (List.rev tempmemorycollector)
			| head::tail ->
			  match (head) with
			      TensorOpt (e1,e2,op_type) -> env_build(tail,
				  (tensor_operation_constructor e1 e2 
				  (string_of_int (List.length tail)) op_type) @ 
				  temporary_declarations,
				  construction,
				  finaldeclaration,tempmemorycollector,1,[DirectOut (e1)],
				  memoryimprint,varname)
				| TensorSub (e1) -> env_build(tail,
				  temporary_declarations, 
				  construction,finaldeclaration,
				tempmemorycollector,1, [TempGetSize (List.length tail)],
				(declare_tensor_in_C_without_first_line (varname) e1),varname)
				| VarDeclare (e1) -> 
				    env_build(tail,
				    temporary_declarations, construction,
				    [DirectOut (e1)] @ 
				    finaldeclaration,tempmemorycollector,declarationtype,
				    declarationsize,memoryimprint,e1)
				| _ -> env_build(tail,temporary_declarations,[head] @ 
				    construction,finaldeclaration,tempmemorycollector,
					declarationtype,declarationsize,memoryimprint,varname)
		  in env_build(listin, [], [], [],[], 0, 
		     [DirectOut "(int[1]){1};\n"],[],"") 
	(* 
	Match the expressions that don't require temporary variable 
	assignments in order to work 
	*)
    in let rec expr = function
	    Literal i -> [Lit i]
	  | GetTensorElement (i,j) -> [DirectOut i]
      | Id s ->
	  (try [DirectOut (List.find (fun x -> x = s) (fdecl.locals))]
            with Not_found -> 
              (try [DirectOut (List.find (fun x -> x = s)  fdecl.formals)]
            with Not_found -> 
              (try [DirectOut (List.find (fun x -> x = s)  globals)]
            with Not_found -> 
              raise (Failure ("TENLAB Error: Undeclared tensor " ^ s)))))
	  |	TensorGet (x) ->  [TensorSub (x)]
      | Binop (e1, op, e2) -> (match e1, e2 with 
          | TensorGet x,TensorGet y -> [TensorOpt (x,y,string_of_stmt (Bin op))]
		  | Literal x,Literal y -> [Lit x]  @ [Bin op] @ [Lit y]
		  | Id x,Literal y -> [DirectOut (x ^ ".Content[" ^ x ^ 
		      ".cur_content_length-1]")] @ [Bin op] @ [Lit y]
		  | Literal x,Id y -> [Lit x] @ [Bin op] @ [DirectOut y] @ 
           (try [DirectOut (".Content[" ^ (List.find (fun d -> d = y)  fdecl.formals)^ ".cur_content_length-1]")]
            with Not_found -> 
              (try [DirectOut (".Content[" ^ (List.find (fun d -> d = y)  globals)^ ".cur_content_length-1]")]
            with Not_found -> 
              raise (Failure ("TENLAB Error: Undeclared tensor " ^ y))))
            (*[DirectOut (y ^ ".Content[" ^ y ^ ".cur_content_length-1]")]*)
		  | Id x,Id y -> [DirectOut x] @ 
           (try [DirectOut (".Content[" ^ (List.find (fun d -> d = x)  fdecl.formals)^ ".cur_content_length-1]")]
            with Not_found -> 
              (try [DirectOut (".Content[" ^ (List.find (fun d -> d = x)  globals)^ ".cur_content_length-1]")]
            with Not_found -> 
              raise (Failure ("TENLAB Error: Undeclared tensor " ^ x)))) @ 
			  [Bin op] @ [DirectOut y] @ 
           (try [DirectOut (".Content[" ^ (List.find (fun d -> d = y)  fdecl.formals)^ ".cur_content_length-1]")]
            with Not_found -> 
              (try [DirectOut (".Content[" ^ (List.find (fun d -> d = y)  globals)^ ".cur_content_length-1]")]
            with Not_found -> 
              raise (Failure ("TENLAB Error: Undeclared tensor " ^ y))))
		  | _, _ -> (expr e1) @ [Bin op] @ (expr e2))
      | Assign (s, e) -> 
	      (try (*let mini_env = [] in *) 
		    let expr_list = expr_to_tensor_op ([VarDeclare s] @ (expr e)) in 
			  (*[DirectOut ((List.find (fun x -> x = s) (fdecl.formals)) ^ 
			  " = ")]*) (expr_list) 
			  with Not_found -> 
			        try [VarAssign (StringMap.find s env.global_index)] @ expr e 
			  with Not_found -> 
			         raise (Failure ("TENLAB Error: Undefined tensor " ^ s))) 
	  | Call ("print", actuals) -> if (List.length actuals) = 1 then
	      [(FunctionCall ("print")) ] @ 
		    (List.concat (List.map expr (List.rev actuals))) @ 
		    [Parend]  (* @ [DirectOut "printf(\"\\n\")"]*) else 
      raise (Failure ("TENLAB Error: function print demands a single input."))
	  | Call ("input", actuals) -> if main_type=0 then 
      raise (Failure ("TENLAB Error: Inputs are only possible in MEX mode."))
      else if (List.length actuals) = 2 then
	      [(FunctionCall ("input")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ","]) [] (List.map expr (List.rev actuals)))))) @ 
          [DirectOut "-1"] @
		      [Parend] else 
      raise (Failure ("TENLAB Error: function input demands two inputs."))
	  | Call ("output", actuals) ->  if main_type=0 then 
      raise (Failure ("TENLAB Error: Outputs are only possible in MEX mode."))
      else if (List.length actuals) = 2 then
	      [(FunctionCall ("output")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ","]) [] (List.map expr (List.rev actuals)))))) @ 
          [DirectOut "-1"] @
		      [Parend] else 
      raise (Failure ("TENLAB Error: function output demands two inputs."))
    | Call ("length", actuals) -> if (List.length actuals) = 1 then
	      [(FunctionCall ("length")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ","]) [] (List.map expr (List.rev actuals)))))) @ 
		  [Parend] else 
      raise (Failure ("TENLAB Error: function length demands a single input."))
	  | Call ("pop", actuals) -> if (List.length actuals) = 1 then
	      [(FunctionCall ("pop")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ","]) [] (List.map expr (List.rev actuals)))))) @ 
		  [Parend] else 
      raise (Failure ("TENLAB Error: function pop demands a single input."))
    | Call ("dequeue", actuals) -> if (List.length actuals) = 1 then
	      [(FunctionCall ("dequeue")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ","]) [] (List.map expr (List.rev actuals)))))) @ 
		  [Parend] else 
      raise (Failure ("TENLAB Error: function dequeue demands a single input."))
    | Call ("reshape", actuals) -> if (List.length actuals) = 2 then
	      [(FunctionCall ("reshape")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ",&"]) [] (List.map expr (List.rev actuals)))))) @ 
		  [Parend] else 
      raise (Failure ("TENLAB Error: function reshape demands two inputs."))
    | Call ("set", actuals) -> if (List.length actuals) = 3 then
	      [(FunctionCall ("set")) ] @ 
		  ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ",&"]) [] (List.map expr (List.rev actuals)))))) @ 
		  [Parend] else 
      raise (Failure ("TENLAB Error: function set demands three inputs."))
    | Call (fname, actuals) -> 
	      
	        [FunctionCall (* (StringMap.find fname env.function_index) *) 
			fname ]  @	   
            (* (List.concat (List.map expr (List.rev actuals))) *)
            ((List.tl(List.rev(List.fold_left(fun i j -> i @
          j @ [DirectOut ","]) [] (List.map expr (List.rev actuals)))))) @ 
		  [Parend] 
    | Noexpr -> []
	  
    in let rec stmt = function
	    Block sl ->  List.concat (List.map stmt sl)
    | Expr e -> expr e @ [Lineend]
    | BuildTensorProd (a,b,c,d,e,f,g,h) -> 
        [DirectOut (create_tensor_product_code a b c d e f g h)]
	  | CreateTensor (s, e) -> [DirectOut (declare_tensor_in_C s e)]
      (* | Return e     -> expr e @ [ReturnValue num_formals] *)
	  | Return e -> [ReturnValue num_formals] @ expr e @ [Lineend]
	  | Clear e -> [ClearValue e] @ [Lineend]
	  | Clean e -> [CleanValue e] @ [Lineend]
    | If (p, t, f) -> let t' = stmt t and f' = stmt f in
	(* expr p @ [Beq(2 + List.length t')] @
	t' @ [Bra(1 + List.length f')] @ f' *)
	      [Beginf] @  expr p @ [Parend] @ [Curlbegin] @ t' @ 
	      [DirectOut "\n}\nelse \n{\n"] @ f' @  [DirectOut "}\n"]
      | For (e1, e2, b) ->	
	  let b' = stmt b in 
	    (*[DirectOut ("TENLAB_Tensor " ^ e1 ^ 
		";\nTENLAB_Tensor_create(&" ^ e1 ^ ");\n")] @ *)
		[DirectOut (declare_for_loop_in_C e1 e2)] @ 
		[DirectOut ("TENLAB_assign(&" ^ e1 ^ "," ^ e2 ^ 
		".Content[TENLAB_for_" ^ e1 ^ "-1]);\n")] @ (b') @ [Curlend] (*@ 
		[DirectOut ("TENLAB_Tensor_destroy(&" ^ e1 ^ ");\n")]*)
	  (*[DirectOut (declare_for_loop_in_C e1 e2)] @ (stmt b) @ [Curlend]*)
      | While (e, b) ->
	  let b' = stmt b and e' = expr e in let out= match (e') with
	      [DirectOut s] -> [DirectOut "while ("] @ [DirectOut s] @ 
		  [DirectOut (".Content[" ^ s ^ ".cur_content_length-1]>0) {\n") ] @
		  b' @ [Curlend]
		| _ -> raise (Failure 
		    ("TENLAB Error: While loops take in a single tensor as argument."))
		in out

	in if (fdecl.fname)="main" then 
	  (* 
	  Generation of C Functions Compatible with the C Library
	  *)
	  if main_type=0 then
	    [FunctionCall 
		(fdecl.fname)] @
		[DirectOut 
		(str_crop_last_char (List.fold_left (fun i j -> i  ^ j ^ ",") 
		"" fdecl.formals))] @[Parend] @ [Curlbegin] @
		(List.fold_left (fun i j -> i @
		[DirectOut ("TENLAB_Tensor " ^ j ^ ";\n" ^
		"TENLAB_Tensor_create(&" ^ j ^ ");\n")]) [] (globals)) @
		stmt (Block fdecl.body) @ [Curlend] else
		(* Generation of Mex Code for C *)
    [DirectOut 
 "void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])"]
   @ [Curlbegin] @ 
		  (List.fold_left (fun i j -> i @ 
		  [DirectOut ("TENLAB_Tensor " ^ j ^ ";\n" ^
		  "TENLAB_Tensor_create(&" ^ j ^ ");\n")]  ) [] (globals)) @
		  stmt (Block fdecl.body) @ [Curlend]
	else 
	  [DirectOut "TENLAB_Tensor "] @
	  [FunctionCall 
	  (fdecl.fname)] @
	  [DirectOut 
	  (str_crop_last_char (List.fold_left 
	  (fun i j -> i  ^ "TENLAB_Tensor " ^ j ^ ",") "" fdecl.formals))] @
	  [Parend] @ [Curlbegin] @
	  (List.fold_left (fun i j -> i @
	  [DirectOut ("TENLAB_Tensor " ^ j ^ ";\n" ^
	  "TENLAB_Tensor_create(&" ^ j ^ ");\n")]) [] (fdecl.locals)) @ 
	  stmt (Block fdecl.body) @ [Curlend] 
	
    in let env = { function_index = function_indexes;
		               global_index = global_indexes;
		               local_index = StringMap.empty } in

  let entry_function_actual = [] in
    
  (*
  Compile the functions 
  *)
  let func_bodies = 
    entry_function_actual :: List.map (translate env) functions in

  { num_globals = List.length globals;
    (* 
    Concatenate the compiled functions and replace the function
       indexes with their actual names 
    *)
    text = Array.of_list (List.map (function
		_ as s -> s) (List.concat (List.rev func_bodies)))
  }
