

%{ open Ccode %}

%token SEMI LPAREN RPAREN LNPAREN RNPAREN LBRACE RBRACE COMMA COLON TENPROD
%token PLUS MINUS TIMES DIVIDE ASSIGN
%token EQ NEQ LT LEQ GT GEQ
%token RETURN IF ELSE FOR WHILE TENSORDEF CLEAR CLEAN FUNCTION END
%token <string> LITERAL
%token <string> ID
%token <string> TENSOR
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE

%start program
%type <Ccode.program> program

%%

program:
  decls EOF { $1}
  | decls main_decl { fst $1, ($2 :: snd $1) }
decls:
   /* nothing */ { [], [] }
 | decls vdecl { ($2 :: fst $1), snd $1 }
 | decls fdecl { fst $1, ($2 :: snd $1) }
 
 
 main_decl:
   stmt_list
     { { fname = "main";
	 formals = [];
	 locals = [];
	 body = List.rev $1 } }

fdecl:
   FUNCTION ID LPAREN formals_opt RPAREN SEMI vdecl_list stmt_list END SEMI
     { { fname = $2;
	 formals = $4;
	 locals = List.rev $7;
	 body = List.rev $8 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    ID                   { [$1] }
  | formal_list COMMA ID { $3 :: $1 }

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
   TENSORDEF ID SEMI { $2 }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI  { Expr($1) }
  | RETURN expr SEMI                                             
      { Return($2) }
  | CLEAR ID SEMI                                                
      { Clear($2) }
  | CLEAN ID SEMI                                                
      { Clean($2) }
  | ID ASSIGN LBRACE num_list RBRACE LBRACE num_list RBRACE ID LBRACE num_list RBRACE 
      TENPROD LBRACE num_list RBRACE ID LBRACE num_list RBRACE SEMI
      { BuildTensorProd( $1, List.rev $4, List.rev $7, $9,
	    List.rev $11, List.rev $15, $17, List.rev $19 ) }
  | IF LPAREN expr RPAREN SEMI stmt_list END SEMI %prec NOELSE        
      { If($3, Block(List.rev $6), Block([])) }
  | IF LPAREN expr RPAREN SEMI stmt_list ELSE SEMI stmt_list END SEMI      
      { If($3, Block(List.rev $6), Block(List.rev $9)) }
  | FOR ID ASSIGN ID SEMI stmt_list END SEMI
      { For($2, $4, Block(List.rev $6)) }
  | WHILE LPAREN expr RPAREN SEMI stmt_list END SEMI  
      { While($3, Block(List.rev $6)) }

expr:
    LITERAL                      { Literal($1) }
  | ID                           { Id($1) }
  | TENSOR                       { TensorGet($1) }
  | expr PLUS   expr             { Binop($1, Add,   $3) }
  | expr MINUS  expr             { Binop($1, Sub,   $3) }
  | expr TIMES  expr             { Binop($1, Mult,  $3) }
  | expr DIVIDE expr             { Binop($1, Div,   $3) }
  | expr EQ     expr             { Binop($1, Equal, $3) }
  | expr NEQ    expr             { Binop($1, Neq,   $3) }
  | expr LT     expr             { Binop($1, Less,  $3) }
  | expr LEQ    expr             { Binop($1, Leq,   $3) }
  | expr GT     expr             { Binop($1, Greater,  $3) }
  | expr GEQ    expr             { Binop($1, Geq,   $3) }
  | ID ASSIGN   expr             { Assign($1, $3) }
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN           { $2 }

actuals_opt:
    /* nothing */                { [] }
  | actuals_list                 { List.rev $1 }

actuals_list:
    expr                         { [$1] }
  | actuals_list COMMA expr      { $3 :: $1 }
  
num_list:
    LITERAL                         { [int_of_string $1] }
  | num_list COMMA LITERAL      { (int_of_string $3) :: $1 }
