# Testlab: The TENLAB Testing Environment for the C Targets
# Author: Mehmet Kerem Turkcan
# Based on the MICROC code

TENLAB="./tenlab"
ulimit -t 30
globallog=testall_c.log
rm -f $globallog
error=0
globalerror=0
keep=0
Usage() {
    exit 1
}
SignalError() {
    if [ $error -eq 0 ] ; then
	echo "FAILED"
	error=1
    fi
    echo "  $1"
}
Compare() {
    generatedfiles="$generatedfiles $3"
    echo diff -b $1 $2 ">" $3 1>&2
    diff -b "$1" "$2" > "$3" 2>&1 || {
	SignalError "$1 differs"
	echo "FAILED $1 differs from $2" 1>&2
    }
}
Run() {
    echo $* 1>&2
    eval $* || {
	SignalError "$1 failed on $*"
	return 1
    }
}
RunFail() {
    echo $* 1>&2
    eval $* && {
	SignalError "failed: $* did not report an error"
	return 1
    }
    return 0
}
Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.c//'`
    reffile=`echo $1 | sed 's/.c$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."
    echo -n "$basename..."
    echo 1>&2
    echo "###### Testing $basename" 1>&2
    generatedfiles=""
    generatedfiles="$generatedfiles ${basename}.c.out" &&
	  Run "gcc ctests/${basename}.c" " -o " ${basename}-c.exe &&
	  Run "./${basename}-c" "> ${basename}.c.out" &&
    Compare ${basename}.c.out ${reffile}.out ${basename}.cc.diff
    if [ $error -eq 0 ] ; then
	  echo "Works!"
	  echo "###### Failed Horribly!" 1>&2
    else
	  echo "###### FAILED" 1>&2
	  globalerror=$error
    fi
}
CheckFail() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.ten//'`
    reffile=`echo $1 | sed 's/.ten$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."
    echo -n "$basename..."
    echo 1>&2
    echo "###### Testing $basename" 1>&2
    generatedfiles=""
    generatedfiles="$generatedfiles ${basename}.err ${basename}.diff" &&
    RunFail "$TENLAB" "-b" "<" $1 "2>" "${basename}.err" ">>" $globallog &&
    Compare ${basename}.err ${reffile}.err ${basename}.diff
    if [ $error -eq 0 ] ; then
	  echo "Works!"
	  echo "###### Failed Horribly!" 1>&2
    else
	  echo "###### FAILED" 1>&2
	  globalerror=$error
    fi
}
shift `expr $OPTIND - 1`
files="ctests/test-*.c"
for file in $files
do
    case $file in
	*test-*)
	    Check $file 2>> $globallog
	    ;;
	*)
	    echo "unknown file type $file"
	    globalerror=1
	    ;;
    esac
done
exit $globalerror
